from typing import Any, Callable

from spice.retry_strategy import Behavior, RetryStrategy
from spice.spice import SpiceCallArgs
from spice.spice_message import SpiceMessages


def default_failure_message(message: str) -> str:
    return f"Failed to convert response for the following reason: {message}\n\nPlease try again."


class ConverterStrategy(RetryStrategy):
    def __init__(
        self,
        converter: Callable[[str], Any],
        retries: int = 0,
        render_failure_message: Callable[[str], str] = default_failure_message,
    ):
        self.converter = converter
        self.retries = retries
        self.render_failure_message = render_failure_message

    def decide(
        self, call_args: SpiceCallArgs, attempt_number: int, model_output: str, name: str
    ) -> tuple[Behavior, SpiceCallArgs, Any, str]:
        try:
            result = self.converter(model_output)
            return Behavior.RETURN, call_args, result, name
        except Exception as e:
            if attempt_number < self.retries:
                messages = SpiceMessages(messages=call_args.messages)
                messages.add_assistant_text(model_output)
                messages.add_user_text(self.render_failure_message(str(e)))
                new_call_args = call_args.model_copy(update={"messages": messages})
                return Behavior.RETRY, new_call_args, None, f"{name}-retry-{attempt_number}-fail"
            else:
                raise ValueError("Failed to get a valid response after all retries")



import asyncio
import os
import sys
from timeit import default_timer as timer

from spice.models import SONNET_3_5, GPT_4o_2024_08_06

# Modify sys.path to ensure the script can run even when it's not part of the installed library.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from spice import Spice


async def speed_compare():
    client = Spice()

    messages = (
        client.new_messages().add_system_text("You are a helpful assistant.").add_user_text("list 100 random words")
    )

    models = [GPT_4o_2024_08_06, SONNET_3_5]
    runs = 3

    for model in models:
        print(f"Timing {model.name}:")
        for i in range(runs):
            start = timer()
            response = await client.get_response(messages=messages, model=model)
            end = timer()
            print(f"    Run {i + 1}: {end - start:.1f}s, {response.characters_per_second:.0f} characters/s")


if __name__ == "__main__":
    asyncio.run(speed_compare())



from __future__ import annotations

import base64
import io
import mimetypes
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Collection,
    ContextManager,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    Union,
    cast,
)

import anthropic
import httpx
import openai
import tiktoken
from anthropic import AsyncAnthropic
from anthropic.types import ImageBlockParam, Message, MessageStreamEvent, TextBlockParam
from openai import AsyncAzureOpenAI, AsyncOpenAI, OpenAI
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionMessageParam,
)
from PIL import Image
from pydantic import BaseModel
from typing_extensions import override

from spice.call_args import SpiceCallArgs
from spice.errors import APIConnectionError, APIError, AuthenticationError, ImageError, InvalidModelError
from spice.spice_message import SpiceMessage

if TYPE_CHECKING:
    from spice.models import Model


# a MessageParam with more constrained structure
class ConstrainedAnthropicMessageParam(TypedDict):
    content: List[Union[TextBlockParam, ImageBlockParam]]
    role: Literal["user", "assistant"]


class TextAndTokens(BaseModel):
    text: Optional[str] = None
    input_tokens: Optional[int] = None
    cache_creation_input_tokens: Optional[int] = None
    cache_read_input_tokens: Optional[int] = None
    output_tokens: Optional[int] = None


class WrappedClient(ABC):
    @abstractmethod
    async def get_chat_completion_or_stream(
        self, call_args: SpiceCallArgs
    ) -> ChatCompletion | AsyncIterator[ChatCompletionChunk] | Message | AsyncIterator[MessageStreamEvent]: ...

    @abstractmethod
    def process_chunk(self, chunk, call_args: SpiceCallArgs) -> TextAndTokens: ...

    @abstractmethod
    def extract_text_and_tokens(self, chat_completion, call_args: SpiceCallArgs) -> TextAndTokens: ...

    @abstractmethod
    def catch_and_convert_errors(self) -> ContextManager[None]: ...

    @abstractmethod
    def count_messages_tokens(self, messages: Collection[SpiceMessage], model: Model | str) -> int:
        """
        Returns the number of tokens used by a prompt if it was sent to an API for a chat completion.
        """

    @abstractmethod
    def count_string_tokens(self, message: str, model: Model | str, full_message: bool) -> int:
        """
        Calculates the tokens in this message. Will not be accurate for a prompt.
        Use count_messages_tokens to get the exact amount of tokens for a prompt.
        If full_message is true, will include the extra 4 tokens used in a chat completion by this message
        if this message is part of a prompt. Do not full_message to true for a response.
        """

    @abstractmethod
    async def get_embeddings(self, input_texts: List[str], model: str) -> List[List[float]]: ...

    @abstractmethod
    def get_embeddings_sync(self, input_texts: List[str], model: str) -> List[List[float]]: ...

    @abstractmethod
    async def get_transcription(self, audio_path: Path, model: str) -> tuple[str, float]: ...


def _spice_message_to_openai_content_part(
    message: SpiceMessage,
) -> Union[ChatCompletionContentPartTextParam, ChatCompletionContentPartImageParam]:
    if message.content.type == "text":
        return {"type": "text", "text": message.content.text}
    elif message.content.type == "image_url":
        return {"type": "image_url", "image_url": {"url": message.content.image_url}}
    else:
        raise ValueError(f"Unknown content type: {message.content.type}")


class WrappedOpenAIClient(WrappedClient):
    def __init__(self, key, base_url=None):
        self._sync_client = OpenAI(api_key=key, base_url=base_url)
        self._client = AsyncOpenAI(api_key=key, base_url=base_url)

    def _convert_messages(self, messages: Collection[SpiceMessage]) -> List[ChatCompletionMessageParam]:
        converted_messages = []
        for message in messages:
            content_part = _spice_message_to_openai_content_part(message)
            if converted_messages and converted_messages[-1]["role"] == message.role:
                converted_messages[-1]["content"].append(content_part)
            else:
                converted_messages.append({"role": message.role, "content": [content_part]})
        return converted_messages

    @override
    async def get_chat_completion_or_stream(self, call_args: SpiceCallArgs):
        # WrappedOpenAIClient can be used with a proxy to a non openai llm, which may not support response_format
        maybe_kwargs: Dict[str, Any] = {}
        if call_args.response_format is not None and "type" in call_args.response_format:
            maybe_kwargs["response_format"] = call_args.response_format
        if call_args.stream:
            maybe_kwargs["stream_options"] = {"include_usage": True}

        # If using vision you have to set max_tokens or api errors
        if call_args.max_tokens is None and "gpt-4" in call_args.model:
            max_tokens = 4096
        else:
            max_tokens = call_args.max_tokens

        converted_messages = self._convert_messages(call_args.messages)

        return await self._client.chat.completions.create(
            model=call_args.model,
            messages=converted_messages,
            stream=call_args.stream,
            temperature=call_args.temperature,
            max_tokens=max_tokens,
            **maybe_kwargs,
        )

    @override
    def process_chunk(self, chunk, call_args: SpiceCallArgs):
        chunk = cast(ChatCompletionChunk, chunk)
        content = None
        if len(chunk.choices) > 0:
            content = chunk.choices[0].delta.content
        input_tokens = None
        output_tokens = None
        if chunk.usage is not None:
            input_tokens = chunk.usage.prompt_tokens
            output_tokens = chunk.usage.completion_tokens
        return TextAndTokens(text=content, input_tokens=input_tokens, output_tokens=output_tokens)

    @override
    def extract_text_and_tokens(self, chat_completion, call_args: SpiceCallArgs):
        return TextAndTokens(
            text=chat_completion.choices[0].message.content,
            input_tokens=chat_completion.usage.prompt_tokens,
            cache_creation_input_tokens=0,
            cache_read_input_tokens=0,
            output_tokens=chat_completion.usage.completion_tokens,
        )

    @override
    @contextmanager
    def catch_and_convert_errors(self):
        # TODO: Do we catch all errors? I think we should catch APIStatusError
        try:
            yield
        except openai.APIConnectionError as e:
            raise APIConnectionError(f"OpenAI Connection Error: {e.message}") from e
        except openai.AuthenticationError as e:
            raise AuthenticationError(f"OpenAI Authentication Error: {e.message}") from e
        except openai.APIStatusError as e:
            raise APIError(f"OpenAI Status Error: {e.message}") from e

    def _get_encoding_for_model(self, model: Model | str) -> tiktoken.Encoding:
        from spice.models import Model

        if isinstance(model, Model):
            model = model.name
        try:
            # OpenAI fine-tuned models are named `ft:<base model>:<name>:<id>`. If tiktoken
            # can't match the full string, it tries to match on startswith, e.g. 'gpt-4'
            _model = model.split(":")[1] if model.startswith("ft:") else model
            return tiktoken.encoding_for_model(_model)
        except KeyError:
            return tiktoken.get_encoding("cl100k_base")

    @override
    def count_messages_tokens(self, messages: Collection[SpiceMessage], model: Model | str) -> int:
        """
        Adapted from https://platform.openai.com/docs/guides/text-generation/managing-tokens
        """

        encoding = self._get_encoding_for_model(model)

        num_tokens = 0
        for message in messages:
            # every message follows <|start|>{role/name}\n{content}<|end|>\n
            # this has 5 tokens (start token, role, \n, end token, \n), but we count the role token later
            num_tokens += 4
            content = message.content
            if content.type == "text":
                num_tokens += len(encoding.encode(content.text))
            elif content.type == "image_url":
                image_base64: str = content.image_url.split(",")[1]
                image_bytes: bytes = base64.b64decode(image_base64)
                image = Image.open(io.BytesIO(image_bytes))
                size = image.size
                # As described here: https://platform.openai.com/docs/guides/vision/calculating-costs
                scale = min(1, 2048 / max(size))
                size = (int(size[0] * scale), int(size[1] * scale))
                scale = min(1, 768 / min(size))
                size = (int(size[0] * scale), int(size[1] * scale))
                num_tokens += 85 + 170 * ((size[0] + 511) // 512) * ((size[1] + 511) // 512)
            else:
                raise ValueError(f"Unknown content type: {content.type}")
        num_tokens += 2  # every reply is primed with <|start|>assistant
        return num_tokens

    @override
    def count_string_tokens(self, message: str, model: Model | str, full_message: bool) -> int:
        encoding = self._get_encoding_for_model(model)
        return len(encoding.encode(message, disallowed_special=())) + (4 if full_message else 0)

    @override
    async def get_embeddings(self, input_texts: List[str], model: str) -> List[List[float]]:
        embeddings = (await self._client.embeddings.create(input=input_texts, model=model)).data
        sorted_embeddings = sorted(embeddings, key=lambda e: e.index)
        return [result.embedding for result in sorted_embeddings]

    @override
    def get_embeddings_sync(self, input_texts: List[str], model: str) -> List[List[float]]:
        embeddings = self._sync_client.embeddings.create(input=input_texts, model=model).data
        sorted_embeddings = sorted(embeddings, key=lambda e: e.index)
        return [result.embedding for result in sorted_embeddings]

    @override
    async def get_transcription(self, audio_path: Path, model: str) -> tuple[str, float]:
        audio_file = open(audio_path, "rb")
        transcript = await self._client.audio.transcriptions.create(
            model=model, file=audio_file, response_format="verbose_json"
        )
        return (transcript.text, transcript.duration)  # pyright: ignore


class WrappedAzureClient(WrappedOpenAIClient):
    def __init__(self, key, endpoint):
        self._client = AsyncAzureOpenAI(
            api_key=key,
            api_version="2023-12-01-preview",
            azure_endpoint=endpoint,
        )

    @override
    def process_chunk(self, chunk, call_args: SpiceCallArgs):
        # In Azure, the first chunk only contains moderation metadata, and an empty choices array
        if not chunk.choices:
            return TextAndTokens()
        else:
            return super().process_chunk(chunk, call_args)


def _spice_message_to_anthropic_block_param(message: SpiceMessage) -> Union[TextBlockParam, ImageBlockParam]:
    if message.content.type == "text":
        block_param = {"type": "text", "text": message.content.text}
    elif message.content.type == "image_url":
        image_url = message.content.image_url
        if image_url.startswith("http"):
            try:
                response = httpx.get(image_url)
            except Exception:
                raise ImageError(f"Error fetching image {image_url}.")
            media_type = response.headers.get("content-type", mimetypes.guess_type(image_url)[0])
            image = base64.b64encode(response.content).decode("utf-8")
        else:
            media_type = image_url.split(";", maxsplit=1)[0].replace("data:", "")
            image = image_url.split(";base64,", maxsplit=1)[1]
        block_param = {
            "type": "image",
            "source": {"type": "base64", "data": image, "media_type": media_type},
        }
    # ignoring types because cache_control is still in beta
    if message.cache:
        block_param["cache_control"] = {"type": "ephemeral"}  # type: ignore
    return block_param  # type: ignore


class WrappedAnthropicClient(WrappedClient):
    def __init__(self, key):
        self._client = AsyncAnthropic(api_key=key)

    def _convert_messages(
        self, messages: Collection[SpiceMessage], add_json_brace: bool
    ) -> Tuple[List[TextBlockParam], List[ConstrainedAnthropicMessageParam]]:
        system_block_params: List[TextBlockParam] = []
        converted_messages: List[ConstrainedAnthropicMessageParam] = []

        for message in messages:
            if message.role == "system":
                if converted_messages:
                    raise ValueError("System messages must be at the start of the conversation.")
                if message.content.type != "text":
                    raise ValueError("System messages must be text.")
                system_block_params.append(_spice_message_to_anthropic_block_param(message))  # type: ignore
            else:
                block_param = _spice_message_to_anthropic_block_param(message)
                if converted_messages and converted_messages[-1]["role"] == message.role:
                    converted_messages[-1]["content"].append(block_param)
                else:
                    converted_messages.append({"role": message.role, "content": [block_param]})

        if add_json_brace:
            if not converted_messages or converted_messages[-1]["role"] != "assistant":
                converted_messages.append({"role": "assistant", "content": []})
            converted_messages[-1]["content"].append({"type": "text", "text": "{"})

        return system_block_params, converted_messages

    @override
    async def get_chat_completion_or_stream(self, call_args: SpiceCallArgs):
        add_json_brace = (
            call_args.response_format is not None and call_args.response_format.get("type", "text") == "json_object"
        )

        # max_tokens is required by anthropic api
        if call_args.max_tokens is None:
            max_tokens = 4096
        else:
            max_tokens = call_args.max_tokens

        # temperature is optional but can't be None
        maybe_temperature_kwargs: dict[str, Any] = (
            {"temperature": call_args.temperature} if call_args.temperature is not None else {}
        )

        system, converted_messages = self._convert_messages(call_args.messages, add_json_brace)

        return await self._client.messages.create(
            model=call_args.model,
            system=system,
            messages=converted_messages,  # type: ignore
            stream=call_args.stream,
            max_tokens=max_tokens,
            extra_headers={"anthropic-beta": "prompt-caching-2024-07-31"},
            **maybe_temperature_kwargs,
        )

    @override
    def process_chunk(self, chunk, call_args: SpiceCallArgs):
        content = None
        input_tokens = None
        cache_creation_input_tokens = None
        cache_read_input_tokens = None
        output_tokens = None
        if chunk.type == "content_block_delta":
            content = chunk.delta.text
        elif chunk.type == "message_start":
            if call_args.response_format is not None and call_args.response_format.get("type") == "json_object":
                content = "{"
            input_tokens = chunk.message.usage.input_tokens
            cache_creation_input_tokens = chunk.message.usage.cache_creation_input_tokens
            cache_read_input_tokens = chunk.message.usage.cache_read_input_tokens
            output_tokens = chunk.message.usage.output_tokens
        elif chunk.type == "message_delta":
            output_tokens = chunk.usage.output_tokens
        return TextAndTokens(
            text=content,
            input_tokens=input_tokens,
            cache_creation_input_tokens=cache_creation_input_tokens,
            cache_read_input_tokens=cache_read_input_tokens,
            output_tokens=output_tokens,
        )

    @override
    def extract_text_and_tokens(self, chat_completion, call_args: SpiceCallArgs):
        add_brace = call_args.response_format is not None and call_args.response_format.get("type") == "json_object"
        return TextAndTokens(
            text=("{" if add_brace else "") + chat_completion.content[0].text,
            input_tokens=chat_completion.usage.input_tokens,
            cache_creation_input_tokens=chat_completion.usage.cache_creation_input_tokens,
            cache_read_input_tokens=chat_completion.usage.cache_read_input_tokens,
            output_tokens=chat_completion.usage.output_tokens,
        )

    @override
    @contextmanager
    def catch_and_convert_errors(self):
        try:
            yield
        except anthropic.APIConnectionError as e:
            raise APIConnectionError(f"Anthropic Connection Error: {e.message}") from e
        except anthropic.AuthenticationError as e:
            raise AuthenticationError(f"Anthropic Authentication Error: {e.message}") from e
        except anthropic.APIStatusError as e:
            raise APIError(f"Anthropic Status Error: {e.message}") from e

    # Anthropic doesn't give us a way to count tokens, so we just use OpenAI's token counting functions and multiply by a pre-determined multiplier
    class _FakeWrappedOpenAIClient(WrappedOpenAIClient):
        def __init__(self):
            pass

    _fake_openai_client = _FakeWrappedOpenAIClient()
    _anthropic_token_multiplier = 1.25

    @override
    def count_messages_tokens(self, messages: Collection[SpiceMessage], model: Model | str) -> int:
        from spice.models import GPT_35_TURBO_0125

        return int(
            self._fake_openai_client.count_messages_tokens(messages, GPT_35_TURBO_0125)
            * self._anthropic_token_multiplier
        )

    @override
    def count_string_tokens(self, message: str, model: Model | str, full_message: bool) -> int:
        from spice.models import GPT_35_TURBO_0125

        return int(
            self._fake_openai_client.count_string_tokens(message, GPT_35_TURBO_0125, full_message)
            * self._anthropic_token_multiplier
        )

    @override
    async def get_embeddings(self, input_texts: List[str], model: str) -> List[List[float]]:
        raise InvalidModelError()

    @override
    def get_embeddings_sync(self, input_texts: List[str], model: str) -> List[List[float]]:
        raise InvalidModelError()

    @override
    async def get_transcription(self, audio_path: Path, model: str) -> tuple[str, float]:
        raise InvalidModelError()



from typing import Any, Callable, Optional

from spice.call_args import SpiceCallArgs
from spice.retry_strategy import Behavior, RetryStrategy, T


class DefaultRetryStrategy(RetryStrategy):
    def __init__(
        self, validator: Optional[Callable[[str], bool]] = None, converter: Callable[[str], T] = str, retries: int = 0
    ):
        self.validator = validator
        self.converter = converter
        self.retries = retries

    def decide(
        self, call_args: SpiceCallArgs, attempt_number: int, model_output: str, name: str
    ) -> tuple[Behavior, SpiceCallArgs, Any, str]:
        if attempt_number == 1 and call_args.temperature is not None:
            call_args = call_args.model_copy(update={"temperature": max(0.2, call_args.temperature)})
        elif attempt_number > 1 and call_args.temperature is not None:
            call_args = call_args.model_copy(update={"temperature": max(0.5, call_args.temperature)})

        if self.validator and not self.validator(model_output):
            if attempt_number < self.retries:
                return Behavior.RETRY, call_args, None, f"{name}-retry-{attempt_number}-fail"
            else:
                raise ValueError("Failed to get a valid response after all retries")
        try:
            result = self.converter(model_output)
            return Behavior.RETURN, call_args, result, name
        except Exception:
            if attempt_number < self.retries:
                return Behavior.RETRY, call_args, None, f"{name}-retry-{attempt_number}-fail"
            else:
                raise ValueError("Failed to get a valid response after all retries")



from __future__ import annotations

import glob
import json
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from json import JSONDecodeError
from pathlib import Path
from timeit import default_timer as timer
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Collection,
    Dict,
    Generic,
    List,
    Optional,
    TypeVar,
    cast,
)

import httpx
from jinja2 import DictLoader, Environment
from openai.types.chat.completion_create_params import ResponseFormat
from pydantic import BaseModel, Field

from spice.call_args import SpiceCallArgs
from spice.errors import InvalidModelError, UnknownModelError
from spice.models import EmbeddingModel, Model, TextModel, TranscriptionModel, get_model_from_name
from spice.providers import Provider, get_provider_from_name
from spice.retry_strategy import Behavior, RetryStrategy
from spice.retry_strategy.default_strategy import DefaultRetryStrategy
from spice.spice_message import SpiceMessage, SpiceMessages
from spice.utils import (
    embeddings_request_cost,
    print_stream,
    string_identity,
    text_request_cost,
    transcription_request_cost,
)
from spice.wrapped_clients import TextAndTokens, WrappedClient

T = TypeVar("T")


class SpiceResponse(BaseModel, Generic[T]):
    call_args: SpiceCallArgs = Field(
        description="""The call arguments given to the model that created this response."""
    )
    text: str = Field(description="""The total text sent by the model.""")
    total_time: float = Field(
        description="""How long it took for the response to be completed.
        May be inaccurate for incomplete streamed responses."""
    )
    input_tokens: int = Field(
        description="""The number of input tokens given in this response.
        May be inaccurate for incomplete streamed responses."""
    )
    cache_creation_input_tokens: int = Field(
        description="""The number of input tokens cached as part of this response."""
    )
    cache_read_input_tokens: int = Field(
        description="""The number of input tokens read from cache as part of this response."""
    )
    output_tokens: int = Field(
        description="""The number of output tokens given by the model in this response.
        May be inaccurate for incomplete streamed responses."""
    )
    completed: bool = Field(
        description="""Whether or not this response was fully completed.
        This will only ever be false for incomplete streamed responses."""
    )
    cost: Optional[float] = Field(
        default=None,
        description="""The cost of this request in cents.
        May be inaccurate for incompleted streamed responses.
        Will be None if the cost of the model used is not known.""",
    )

    def __init__(self, result: Optional[T] = None, **kwargs):
        super().__init__(**kwargs)
        self._result = result

    @property
    def result(self) -> T:
        """The result of this response, if a converter was given, otherwise the raw text."""
        if self._result is None:
            return self.text  # type: ignore
        return self._result

    @property
    def total_tokens(self) -> int:
        """The total tokens, input and output, in this response."""
        return self.input_tokens + self.output_tokens

    @property
    def characters_per_second(self) -> float:
        """
        The characters per second that the model output.
        May be inaccurate for streamed responses if not iterated over and completed immediately.
        """
        return len(self.text) / self.total_time


class StreamingSpiceResponse:
    """
    Returned from a streaming llm call. Can be iterated over asynchronously to retrieve the content.
    """

    def __init__(
        self,
        model: TextModel,
        call_args: SpiceCallArgs,
        client: WrappedClient,
        stream: AsyncIterator,
        callback: Optional[Callable[[SpiceResponse], None]] = None,
        streaming_callback: Optional[Callable[[str], None]] = None,
    ):
        self._model = model
        self._call_args = call_args
        self._text = []
        self._start_time = timer()
        self._end_time = None
        self._input_tokens = None
        self._cache_creation_input_tokens = None
        self._cache_read_input_tokens = None
        self._output_tokens = None
        self._finished = False
        self._client = client
        self._stream = stream
        self._callback = callback
        self._streaming_callback = streaming_callback

    def __aiter__(self):
        return self

    async def __anext__(self):
        with self._client.catch_and_convert_errors():
            try:
                content = None
                while content is None:
                    chunk = await anext(self._stream)
                    text_and_tokens = self._client.process_chunk(chunk, self._call_args)
                    content = text_and_tokens.text
                    if text_and_tokens.input_tokens is not None:
                        self._input_tokens = text_and_tokens.input_tokens
                    if text_and_tokens.cache_creation_input_tokens is not None:
                        self._cache_creation_input_tokens = text_and_tokens.cache_creation_input_tokens
                    if text_and_tokens.cache_read_input_tokens is not None:
                        self._cache_read_input_tokens = text_and_tokens.cache_read_input_tokens
                    if text_and_tokens.output_tokens is not None:
                        self._output_tokens = text_and_tokens.output_tokens
                if self._streaming_callback is not None:
                    self._streaming_callback(content)
                self._text.append(content)
                return content
            except StopAsyncIteration:
                self._end_time = timer()
                self._finished = True
                raise

    def current_response(self) -> SpiceResponse:
        """
        Returns a SpiceResponse containing the response as it's been received so far.
        Will not wait for the LLM call to finish.
        """
        if not self._finished or self._end_time is None:
            self._end_time = timer()

        full_output = "".join(self._text)

        input_tokens = self._input_tokens
        if input_tokens is None:
            input_tokens = self._client.count_messages_tokens(self._call_args.messages, self._model)
        cache_creation_input_tokens = self._cache_creation_input_tokens
        if cache_creation_input_tokens is None:
            cache_creation_input_tokens = 0
        cache_read_input_tokens = self._cache_read_input_tokens
        if cache_read_input_tokens is None:
            cache_read_input_tokens = 0
        output_tokens = self._output_tokens
        if output_tokens is None:
            output_tokens = self._client.count_string_tokens(full_output, self._model, full_message=False)

        cost = text_request_cost(
            self._model, input_tokens, cache_creation_input_tokens, cache_read_input_tokens, output_tokens
        )
        response = SpiceResponse(
            call_args=self._call_args,
            text=full_output,
            total_time=self._end_time - self._start_time,
            input_tokens=input_tokens,
            cache_creation_input_tokens=cache_creation_input_tokens,
            cache_read_input_tokens=cache_read_input_tokens,
            output_tokens=output_tokens,
            completed=self._finished,
            cost=cost,
        )
        if self._callback is not None:
            self._callback(response)

        return response

    async def complete_response(self) -> SpiceResponse:
        """
        Waits until the entire LLM call finishes, collects it, and returns its SpiceResponse.
        """
        async for _ in self:
            pass
        return self.current_response()


@dataclass
class EmbeddingResponse:
    embeddings: List[List[float]]
    """The list of embeddings of the list of input texts."""

    total_time: float
    """How long it took for the response to be completed."""

    input_tokens: int
    """The number of input tokens given in this response."""

    cost: Optional[float]
    """The cost of this request in cents. Will be None if the cost of the model used is not known."""


@dataclass
class TranscriptionResponse:
    text: str
    """The transcription of the input audio."""

    total_time: float
    """How long it took for the response to be completed."""

    input_length: float
    """The length of the input audio in seconds."""

    cost: Optional[float]
    """The cost of this request in cents. Will be None if the cost of the model used is not known."""


class Spice:
    """
    The Spice client. The majority of the time, only one Spice client should be initialized and used.
    Automatically handles multiple providers and their respective errors.
    """

    def __init__(
        self,
        default_text_model: Optional[TextModel | str] = None,
        default_embeddings_model: Optional[EmbeddingModel | str] = None,
        model_aliases: Optional[Dict[str, Model | str]] = None,
        logging_dir: Optional[Path | str] = None,
        logging_callback: Optional[Callable[[SpiceResponse, str, str], None]] = None,
        default_temperature: Optional[float] = None,
    ):
        """
        Creates a new Spice client.

        Args:
            default_text_model: The default model that will be used for chat completions if no other model is given.
            Will raise an InvalidModelError if the model is not a text model.

            default_embeddings_model: The default model that will be used for embeddings if no other model is given.
            Will raise an InvalidModelError if the model is not an embeddings model.

            model_aliases: A custom model name map.

            logging_dir: If not None, will log all api calls to the given directory.

            logging_callback: If not None, will call the given function with the SpiceResponse, the name of the run,
            and the name of the call after every call finishes.

            default_temperature: The default temperature to use for chat completions if no other temperature is given.
        """

        if isinstance(default_text_model, str):
            text_model = get_model_from_name(default_text_model)
        else:
            text_model = default_text_model
        if text_model and not isinstance(text_model, TextModel):
            raise InvalidModelError("Default text model must be a text model")
        self._default_text_model = text_model

        if isinstance(default_embeddings_model, str):
            embeddings_model = get_model_from_name(default_embeddings_model)
        else:
            embeddings_model = default_embeddings_model
        if embeddings_model and not isinstance(embeddings_model, EmbeddingModel):
            raise InvalidModelError("Default embeddings model must be an embeddings model")
        self._default_embeddings_model = embeddings_model
        self._default_temperature = default_temperature

        # TODO: Should we validate model aliases?
        self._model_aliases = model_aliases

        self._total_cost: float = 0
        self._prompts: Dict[str, str] = {}

        self.logging_dir = None if logging_dir is None else Path(logging_dir).expanduser()
        self.logging_callback = logging_callback
        self.new_run("spice")

    def new_messages(self) -> SpiceMessages:
        """
        Returns a new SpiceMessages object.
        """
        return SpiceMessages(self)

    def new_run(self, name: str):
        """
        Create a new run. All llm calls will be logged in a folder with the run name and a timestamp.
        """
        timestamp = datetime.now().strftime("%m%d%y_%H%M%S")
        self._cur_run = f"{name}_{timestamp}"
        self._cur_logged_names = defaultdict(int)
        self._log_prompts()

    def _log_prompts(self):
        if self.logging_dir is not None:
            logging_dir = self.logging_dir / self._cur_run
            logging_dir.mkdir(exist_ok=True, parents=True)
            with open(logging_dir / "prompts.json", "w") as file:
                json.dump(self._prompts, file)

    def _log_response(self, response: SpiceResponse, name: Optional[str] = None):
        base_name = "spice" if name is None else name

        if self.logging_dir is not None:
            full_name = f"{base_name}_{self._cur_logged_names[base_name]}.json"
            self._cur_logged_names[base_name] += 1
            response_json = response.model_dump_json()

            logging_dir = self.logging_dir / self._cur_run
            logging_dir.mkdir(exist_ok=True, parents=True)
            with (logging_dir / full_name).open("w") as file:
                file.write(f"{response_json}\n")
        if self.logging_callback is not None:
            self.logging_callback(response, self._cur_run, base_name)

    @property
    def total_cost(self) -> float:
        """The total cost in cents of all api calls made through this Spice client."""
        return self._total_cost

    def load_provider(self, provider: Provider | str):
        """
        Loads the specified provider and raises a NoAPIKeyError if no valid api key for the provider is found.
        Providers not preloaded will be loaded on first use, and the NoAPIKeyError will be raised when they are used.
        """

        if isinstance(provider, str):
            provider = get_provider_from_name(provider)
        provider.get_client()

    def _get_client(self, model: Model, provider: Optional[Provider | str]) -> WrappedClient:
        if provider is not None:
            if isinstance(provider, str):
                provider = get_provider_from_name(provider)
            return provider.get_client()
        else:
            if model.provider is None:
                raise UnknownModelError("Provider is required when unknown models are used")
            return model.provider.get_client()

    def _get_model(self, model: Model | str) -> Model:
        if self._model_aliases is not None and model in self._model_aliases:
            model = self._model_aliases[model]

        if isinstance(model, str):
            model = get_model_from_name(model)

        return model

    def _get_text_model(self, model: Optional[Model | str]) -> TextModel:
        if model is None:
            if self._default_text_model is None:
                raise InvalidModelError("Model is required when default text model is not set at initialization")
            model = self._default_text_model

        if self._model_aliases is not None and model in self._model_aliases:
            model = self._model_aliases[model]

        if isinstance(model, str):
            model = get_model_from_name(model)

        if not isinstance(model, TextModel):
            raise InvalidModelError(f"Model {model} is not a text model")

        return model

    def _fix_call_args(
        self,
        messages: List[SpiceMessage],
        model: Model,
        stream: bool,
        temperature: Optional[float],
        max_tokens: Optional[int],
        response_format: Optional[ResponseFormat],
    ):
        # Not all providers support response format
        if response_format is not None:
            if response_format == {"type": "text"}:
                response_format = None

        return SpiceCallArgs(
            model=model.name,
            messages=list(messages),  # convert from SpiceMessages so we can serialize
            stream=stream,
            temperature=self._default_temperature if temperature is None else temperature,
            max_tokens=max_tokens,
            response_format=response_format,
        )

    async def get_response(
        self,
        messages: List[SpiceMessage],
        model: Optional[TextModel | str] = None,
        provider: Optional[Provider | str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        response_format: Optional[ResponseFormat] = None,
        name: Optional[str] = None,
        validator: Optional[Callable[[str], bool]] = None,
        converter: Callable[[str], T] = string_identity,
        streaming_callback: Optional[Callable[[str], None]] = None,
        retries: int = 0,
        retry_strategy: Optional[RetryStrategy[T]] = None,
    ) -> SpiceResponse[T]:
        """
        Asynchronously retrieves a chat completion response.

        Args:
            messages: The list of messages given as context for the completion.
            Will raise an ImageError if any invalid images are given.

            model: The model to use. Must be a text based model. If no model is given, will use the default text model
            the client was initialized with. If the model is unknown to Spice, a provider must be given.
            Will raise an InvalidModelError if the model is not a text model.
            Will raise an UnknownModelError if the model is unknown and no provider was given.

            provider: The provider to use. If specified, will override the model's default provider if known.
            Must be specified if an unknown model is used.

            temperature: The temperature to give the model.

            max_tokens: The maximum tokens the model can output.

            response_format: For valid models, will set the response format to 'text' or 'json'.
            If the provider/model does not support response_format, this argument will be ignored.

            name: If given, will be given this name when logged.

            validator: If given, will be called with the text of the response.
            If it returns False, the response will be discarded and another attempt will be made.

            converter: If given, will be called with the text of the response.
            The result of the converter will be the result of the response.
            If the converter throws an exception, the response will be discarded
            and another attempt will be made up to retries times.

            streaming_callback: If given, will be called with the text of the response as it is received.

            retries: The number of times to retry getting a valid response. If 0, will not retry.

            retry_strategy: The strategy to use for retrying the request. If None, a DefaultRetryStrategy will be used.
            The strategy determines which model responses will be accepted and on an invalid response how the call_args
            will be modified.
        """
        if retry_strategy is None:
            retry_strategy = DefaultRetryStrategy(validator, converter, retries)

        cost = 0
        attempt_number = 0
        text_model = self._get_text_model(model)
        call_args = self._fix_call_args(
            messages, text_model, streaming_callback is not None, temperature, max_tokens, response_format
        )
        while True:
            start_time = timer()
            text_model = self._get_text_model(call_args.model)
            client = self._get_client(text_model, provider)

            with client.catch_and_convert_errors():
                if streaming_callback is not None:
                    stream = await client.get_chat_completion_or_stream(call_args)
                    stream = cast(AsyncIterator, stream)
                    streaming_spice_response = StreamingSpiceResponse(
                        text_model, call_args, client, stream, None, streaming_callback
                    )
                    chat_completion = await streaming_spice_response.complete_response()
                    if streaming_callback == print_stream:
                        print()
                    text_and_tokens = TextAndTokens(
                        text=chat_completion.text,
                        input_tokens=chat_completion.input_tokens,
                        cache_creation_input_tokens=chat_completion.cache_creation_input_tokens,
                        cache_read_input_tokens=chat_completion.cache_read_input_tokens,
                        output_tokens=chat_completion.output_tokens,
                    )
                else:
                    chat_completion = await client.get_chat_completion_or_stream(call_args)
                    text_and_tokens = client.extract_text_and_tokens(chat_completion, call_args)

            completion_cost = text_request_cost(
                text_model,
                text_and_tokens.input_tokens,  # type: ignore
                text_and_tokens.cache_creation_input_tokens,  # type: ignore
                text_and_tokens.cache_read_input_tokens,  # type: ignore
                text_and_tokens.output_tokens,  # type: ignore
            )
            if completion_cost is not None:
                cost += completion_cost
                self._total_cost += completion_cost

            end_time = timer()

            behavior, next_call_args, result, call_name = retry_strategy.decide(
                call_args,
                attempt_number,
                text_and_tokens.text,  # type: ignore
                name or "",
            )
            response = SpiceResponse(
                call_args=call_args,
                text=text_and_tokens.text,  # type: ignore
                total_time=end_time - start_time,
                input_tokens=text_and_tokens.input_tokens,  # type: ignore
                cache_creation_input_tokens=text_and_tokens.cache_creation_input_tokens,  # type: ignore
                cache_read_input_tokens=text_and_tokens.cache_read_input_tokens,  # type: ignore
                output_tokens=text_and_tokens.output_tokens,  # type: ignore
                completed=True,
                cost=cost,
                result=result,
            )
            self._log_response(response, call_name)
            if behavior == Behavior.RETURN:
                return response
            else:
                attempt_number += 1
                call_args = next_call_args

    async def stream_response(
        self,
        messages: List[SpiceMessage],
        model: Optional[TextModel | str] = None,
        provider: Optional[Provider | str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        response_format: Optional[ResponseFormat] = None,
        name: Optional[str] = None,
        streaming_callback: Optional[Callable[[str], None]] = None,
    ) -> StreamingSpiceResponse:
        """
        Asynchronously retrieves a chat completion stream that can be iterated over asynchronously.

        Args:
            messages: The list of messages given as context for the completion.
            Will raise an ImageError if any invalid images are given.

            model: The model to use. Must be a text based model. If no model is given, will use the default text model
            the client was initialized with. If the model is unknown to Spice, a provider must be given.
            Will raise an InvalidModelError if the model is not a text model.
            Will raise an UnknownModelError if the model is unknown and no provider was given.

            provider: The provider to use. If specified, will override the model's default provider if known.
            Must be specified if an unknown model is used.

            temperature: The temperature to give the model.

            max_tokens: The maximum tokens the model can output.

            response_format: For valid models, will set the response format to 'text' or 'json'.
            If the provider/model does not support response_format, this argument will be ignored.

            name: If given, will be given this name when logged.
        """
        text_model = self._get_text_model(model)
        client = self._get_client(text_model, provider)
        call_args = self._fix_call_args(messages, text_model, True, temperature, max_tokens, response_format)

        with client.catch_and_convert_errors():
            stream = await client.get_chat_completion_or_stream(call_args)
        stream = cast(AsyncIterator, stream)

        def callback(response: SpiceResponse, cache: List[float] = [0]):
            if response.cost is not None:
                self._total_cost += response.cost - cache[0]
                cache[0] = response.cost
            # TODO: Do we want to log multiple times? Our old log might be incomplete, which is why this is still in,
            # but probably not necessary.
            self._log_response(response, name)

        return StreamingSpiceResponse(text_model, call_args, client, stream, callback, streaming_callback)

    def _get_embedding_model(self, model: Optional[Model | str]) -> EmbeddingModel:
        if model is None:
            if self._default_embeddings_model is None:
                raise InvalidModelError("Model is required when default embeddings model is not set at initialization")
            model = self._default_embeddings_model

        if self._model_aliases is not None and model in self._model_aliases:
            model = self._model_aliases[model]

        if isinstance(model, str):
            model = get_model_from_name(model)

        if not isinstance(model, EmbeddingModel):
            raise InvalidModelError(f"Model {model} is not a embedding model")

        return model

    async def get_embeddings(
        self,
        input_texts: List[str],
        model: Optional[EmbeddingModel | str] = None,
        provider: Optional[Provider | str] = None,
    ) -> EmbeddingResponse:
        """
        Asynchronously retrieves embeddings for a list of text.

        Args:
            input_texts: The texts to generate embeddings for.

            model: The embedding model to use. If no model is given, will use the default embedding model
            the client was initialized with. If the model is unknown to Spice, a provider must be given.
            Will raise an InvalidModelError if the model is not a embedding model.
            Will raise an UnknownModelError if the model is unknown and no provider was given.

            provider: The provider to use. If specified, will override the model's default provider if known.
            Must be specified if an unknown model is used.
        """

        embedding_model = self._get_embedding_model(model)
        client = self._get_client(embedding_model, provider)

        input_tokens = sum(client.count_string_tokens(text, embedding_model, False) for text in input_texts)
        cost = embeddings_request_cost(embedding_model, input_tokens)
        if cost is not None:
            self._total_cost += cost

        start_time = timer()
        with client.catch_and_convert_errors():
            embeddings = await client.get_embeddings(input_texts, embedding_model.name)
        end_time = timer()

        return EmbeddingResponse(embeddings, end_time - start_time, input_tokens, cost)

    def get_embeddings_sync(
        self,
        input_texts: List[str],
        model: Optional[EmbeddingModel | str] = None,
        provider: Optional[Provider | str] = None,
    ) -> EmbeddingResponse:
        """
        Synchronously retrieves embeddings for a list of text.

        Args:
            input_texts: The texts to generate embeddings for.

            model: The embedding model to use. If no model is given, will use the default embedding model
            the client was initialized with. If the model is unknown to Spice, a provider must be given.
            Will raise an InvalidModelError if the model is not a embedding model.
            Will raise an UnknownModelError if the model is unknown and no provider was given.

            provider: The provider to use. If specified, will override the model's default provider if known.
            Must be specified if an unknown model is used.
        """

        embedding_model = self._get_embedding_model(model)
        client = self._get_client(embedding_model, provider)

        input_tokens = sum(client.count_string_tokens(text, embedding_model, False) for text in input_texts)
        cost = embeddings_request_cost(embedding_model, input_tokens)
        if cost is not None:
            self._total_cost += cost

        start_time = timer()
        with client.catch_and_convert_errors():
            embeddings = client.get_embeddings_sync(input_texts, embedding_model.name)
        end_time = timer()

        return EmbeddingResponse(embeddings, end_time - start_time, input_tokens, cost)

    def _get_transcription_model(self, model: Model | str) -> TranscriptionModel:
        if self._model_aliases is not None and model in self._model_aliases:
            model = self._model_aliases[model]

        if isinstance(model, str):
            model = get_model_from_name(model)

        if not isinstance(model, TranscriptionModel):
            raise InvalidModelError(f"Model {model} is not a transcription model")

        return model

    async def get_transcription(
        self,
        audio_path: Path | str,
        model: TranscriptionModel | str,
        provider: Optional[Provider | str] = None,
    ) -> TranscriptionResponse:
        """
        Asynchronously retrieves embeddings for a list of text.

        Args:
            audio_path: The path to the audio file to transcribe.

            model: The model to use. Must be a transciption model.
            If the model is unknown to Spice, a provider must be given.
            Will raise an InvalidModelError if the model is not a transciption model.
            Will raise an UnknownModelError if the model is unknown and no provider was given.

            provider: The provider to use. If specified, will override the model's default provider if known.
            Must be specified if an unknown model is used.
        """

        transcription_model = self._get_transcription_model(model)
        client = self._get_client(transcription_model, provider)

        start_time = timer()
        with client.catch_and_convert_errors():
            transcription, input_length = await client.get_transcription(
                Path(audio_path).expanduser().resolve(), transcription_model.name
            )
        end_time = timer()

        cost = transcription_request_cost(transcription_model, input_length)
        if cost is not None:
            self._total_cost += cost

        return TranscriptionResponse(transcription, end_time - start_time, input_length, cost)

    def count_tokens(
        self, text: str, model: Model | str, provider: Optional[Provider | str] = None, is_message: bool = False
    ) -> int:
        """
        Calculates the tokens in the given text. Will not be accurate for a chat completion prompt.
        Use count_prompt_tokens to get the exact amount of tokens for a prompt.

        Args:
            text: The text to count the tokens of.

            model: The model whose tokenizer will be used.
            Will raise UnknownModelError if the model is unknown and no provider is given.

            provider: The provider to use. If specified, will override the model's default provider if known.
            Must be specified if an unknown model is used.

            is_message: If true, will include the extra tokens that messages in chat completions add on.
            Most of the time, you'll want to keep this false.
        """

        model = self._get_model(model)
        client = self._get_client(model, provider)

        return client.count_string_tokens(text, model, is_message)

    def count_prompt_tokens(
        self, messages: Collection[SpiceMessage], model: Model | str, provider: Optional[Provider | str] = None
    ) -> int:
        """
        Calculates the tokens that the messages would have if used in a chat completion.

        Args:
            messages: The messages to count the tokens of.

            model: The model whose tokenizer will be used.
            Will raise UnknownModelError if the model is unknown and no provider is given.

            provider: The provider to use. If specified, will override the model's default provider if known.
            Must be specified if an unknown model is used.
        """

        model = self._get_model(model)
        client = self._get_client(model, provider)

        return client.count_messages_tokens(messages, model)

    ### Prompts ###

    _prompts_dirty = False

    def store_prompt(self, prompt: str, name: str):
        """
        Stores a prompt under the given name

        Args:
            prompt: The prompt to store.

            name: The name of the prompt.
            If the name collides with the name of a previously loaded prompt, the previous prompt will be overwritten.
        """

        if name in self._prompts:
            # Overwriting prompt; should we do anything?
            pass
        self._prompts[name] = prompt
        self._prompts_dirty = True

    def get_prompt(self, name: str) -> str:
        """
        Gets the prompt currently mapped to the given name.

        Args:
            name: The name of the prompt.
        """

        if self._prompts_dirty:
            self._log_prompts()
            self._prompts_dirty = False

        return self._prompts[name]

    def get_rendered_prompt(self, name: str, **context: Any) -> str:
        """
        Gets the prompt currently mapped to the given name and renders it using jinja.

        Args:
            name: The name of the prompt.

            context: The jinja kwargs to render the prompt with.
        """

        if self._prompts_dirty:
            self._log_prompts()
            self._prompts_dirty = False

        return Environment(loader=DictLoader(self._prompts)).get_template(name).render(context)

    def load_prompt(self, file_path: Path | str, name: Optional[str] = None):
        """
        Loads a prompt from the given file.

        Args:
            file_path: The path to the file. Must be a text encoded file.

            name: The name of the prompt. If no name is given, the name will be the file name without the extension;
            i.e., the name of /path/to/prompt.txt will be `prompt`.
            If the name collides with the name of a previously loaded prompt, the previous prompt will be overwritten.
        """

        file_path = Path(file_path).expanduser().resolve()

        try:
            prompt = file_path.read_text()
        except (UnicodeDecodeError, FileNotFoundError):
            raise

        if name is None:
            name = file_path.name.rsplit(".", 1)[0]

        self.store_prompt(prompt, name)

    def _load_toml_dict(self, toml_dict: Dict[str, Any], name: str):
        for k, v in toml_dict.items():
            new_name = f"{name}.{k}"
            if isinstance(v, str):
                self.store_prompt(v, new_name)
            elif isinstance(v, Dict):
                self._load_toml_dict(v, new_name)
            else:
                raise ValueError("Invalid prompt TOML: TOML must only contain strings for prompts.")

    def load_toml_prompts(self, file_path: Path | str, name: Optional[str] = None):
        """
        Loads prompts from a given toml file. The names of the prompts in the file will be prefixed by `{name}.`,
        so a file named prompts.toml containing `prompt1 = "Hi!"` would load the prompt as `prompts.prompt1`.
        Only available for Python 3.11 (when tomllib was introduced)!

        Args:
            file_path: The path to the file. Must be a text encoded file containing valid TOML.

            name: The name of the prompt. If no name is given, the name will be the file name without the extension;
            i.e., the name of /path/to/prompt.toml will be `prompt`.
            If the name collides with the name of a previously loaded prompt, the previous prompt will be overwritten.
        """

        import tomllib

        file_path = Path(file_path).expanduser().resolve()

        try:
            toml = file_path.read_text()
        except (UnicodeDecodeError, FileNotFoundError):
            raise

        if name is None:
            name = file_path.name.rsplit(".", 1)[0]

        self._load_toml_dict(tomllib.loads(toml), name)

    def load_dir(self, dir_path: Path | str):
        """
        Loads a prompts from a given directory.
        Will recursively load all text encoded .txt (and .toml if using Python 3.11) files
        in the directory and its subdirectories.
        Prompt names will be the filenames with their extension stripped with a `.` separating each subdirectory.
        For example, this directory would have these prompt names:

        ```
        dir_path/
            sub_dir/
                prompt_1.toml        - sub_dir.prompt_1.{TOML prompt names}
                not_a_prompt.jpg
            prompt_2.txt            - prompt_2
            another_prompt.txt      - another_prompt.txt
        ```
        If any name collides with the name of a previously loaded prompt, the previous prompt will be overwritten.

        Args:
            dir_path: The path to the directory.
        """

        try:
            import tomllib

            use_toml = True
        except ImportError:
            use_toml = False

        dir_path = Path(dir_path).expanduser().resolve()
        if not dir_path.exists():
            raise FileNotFoundError()

        file_paths = glob.glob(f"{dir_path}/**/*.txt", recursive=True) + glob.glob(
            f"{dir_path}/**/*.toml", recursive=True
        )
        for file_path in file_paths:
            file_path = Path(file_path)

            try:
                prompt = file_path.read_text()
            except (UnicodeDecodeError, FileNotFoundError):
                continue

            rel_path = file_path.relative_to(dir_path)
            name = ".".join(
                (part if i != len(rel_path.parts) - 1 else part.rsplit(".", 1)[0])
                for i, part in enumerate(rel_path.parts)
            )
            if str(file_path).endswith(".toml"):
                if use_toml:
                    self._load_toml_dict(tomllib.loads(prompt), name)  # pyright: ignore[reportPossiblyUnboundVariable]
            elif str(file_path).endswith(".txt"):
                self.store_prompt(prompt, name)

    def load_url(self, url: str, name: str):
        """
        Loads a prompt from the given url. Will send a GET request to the url and expects a response with either the
        following schema or a toml file (.toml only available for Python 3.11 and onwards):
        ```
        {
            "prompts": [
                {
                    "name": "<prompt_name>",
                    "prompt": "<prompt>"
                }, ...
            ]
        }
        ```
        TOML will only be loaded if the url points to a .toml file.
        All prompt names will be prefixed with `name.`
        If any name collides with the name of a previously loaded prompt, the previous prompt will be overwritten.

        Args:
            url: The url.

            name: The name to prefix the prompts with.
        """

        response = httpx.get(url).content.decode("utf-8")
        if url.endswith(".toml"):
            import tomllib

            try:
                prompts = tomllib.loads(response)
            except tomllib.TOMLDecodeError:
                raise
            self._load_toml_dict(prompts, name)

        else:
            try:
                prompts = json.loads(response)
            except JSONDecodeError:
                raise

            for prompt_object in prompts["prompts"]:
                name, prompt = prompt_object["name"], prompt_object["prompt"]
                self.store_prompt(prompt, name)



import asyncio
import os
import random
import sys

import requests
from termcolor import cprint

from spice import Spice
from spice.models import SONNET_3_5
from spice.utils import print_stream

# Modify sys.path to ensure the script can run even when it's not part of the installed library.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))


def display_response_stats(response):
    cprint(f"Input tokens: {response.input_tokens}", "cyan")
    cprint(f"Cache Creation Input Tokens: {response.cache_creation_input_tokens}", "cyan")
    cprint(f"Cache Read Input Tokens: {response.cache_read_input_tokens}", "cyan")
    cprint(f"Output tokens: {response.output_tokens}", "cyan")
    cprint(f"Total time: {response.total_time:.2f}s", "green")
    cprint(f"Cost: ${response.cost / 100:.2f}", "green")


async def run(cache: bool):
    cprint(f"Caching on: {cache}", "cyan")
    client = Spice()
    model = SONNET_3_5
    book_text = requests.get("https://www.gutenberg.org/cache/epub/42671/pg42671.txt").text

    messages = (
        client.new_messages()
        .add_system_text(f"Answer questions about a book. Seed:{random.random()}")
        .add_user_text(f"<book>{book_text}</book>", cache=cache)
    )

    cprint("First model response:", "green")
    response = await client.get_response(
        messages=messages.copy().add_user_text("how many chapters are there? no elaboration."),
        model=model,
        streaming_callback=print_stream,
    )
    display_response_stats(response)

    cprint("Second model response:", "green")
    response = await client.get_response(
        messages=messages.copy().add_user_text("how many volumes are there? no elaboration."),
        model=model,
        streaming_callback=print_stream,
    )
    display_response_stats(response)


if __name__ == "__main__":
    cache = any(arg in sys.argv for arg in ["--cache", "-c"])
    asyncio.run(run(cache))



from typing import List, Optional

from openai.types.chat.completion_create_params import ResponseFormat
from pydantic import BaseModel

from spice.spice_message import SpiceMessage


class SpiceCallArgs(BaseModel):
    model: str
    messages: List[SpiceMessage]
    stream: bool = False
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None
    response_format: Optional[ResponseFormat] = None



from __future__ import annotations

import base64
import mimetypes
from collections.abc import Collection
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from spice.errors import ImageError

if TYPE_CHECKING:
    from spice.spice import Spice

VALID_MIMETYPES = ["image/jpeg", "image/png", "image/gif", "image/webp"]

Role = Literal["user", "assistant", "system"]


class ImageContent(BaseModel):
    type: Literal["image_url"]
    image_url: str


class TextContent(BaseModel):
    type: Literal["text"]
    text: str


class PromptMetadata(BaseModel):
    name: str
    content: str
    context: Dict[str, Any]


class SpiceMessage(BaseModel):
    role: Role
    content: Union[TextContent, ImageContent]
    cache: bool = Field(default=False)
    prompt_metadata: Optional[PromptMetadata]

    def __init__(
        self,
        role: Role,
        text: Optional[str] = None,
        image_url: Optional[str] = None,
        cache: bool = False,
        prompt_metadata: Optional[PromptMetadata] = None,
    ):
        if text is not None and image_url is not None:
            raise ValueError("Only one of text or image_url can be provided.")
        if text is not None:
            content = TextContent(type="text", text=text)
        elif image_url is not None:
            content = ImageContent(type="image_url", image_url=image_url)
        else:
            raise ValueError("Either text or image_url must be provided.")
        super().__init__(role=role, content=content, cache=cache, prompt_metadata=prompt_metadata)


class SpiceMessages(List[SpiceMessage]):
    """A collection of messages to be sent to an API endpoint."""

    def __init__(self, client: Optional[Spice] = None, messages: Collection[SpiceMessage] = []):
        self._client = client
        super().__init__(messages)

    def add_text(self, role: Role, text: str, cache: bool = False) -> SpiceMessages:
        self.append(SpiceMessage(role=role, text=text, cache=cache))
        return self

    def add_user_text(self, text: str, cache: bool = False) -> SpiceMessages:
        return self.add_text("user", text, cache)

    def add_system_text(self, text: str, cache: bool = False) -> SpiceMessages:
        return self.add_text("system", text, cache)

    def add_assistant_text(self, text: str, cache: bool = False) -> SpiceMessages:
        return self.add_text("assistant", text, cache)

    def add_user_image_from_url(self, url: str, cache: bool = False) -> SpiceMessages:
        if not (url.startswith("http://") or url.startswith("https://")):
            raise ImageError(f"Invalid image URL {url}: Must be http or https protocol.")
        self.append(SpiceMessage(role="user", image_url=url, cache=cache))
        return self

    def add_user_image_from_file(self, file_path: Path | str, cache: bool = False) -> SpiceMessages:
        file_path = Path(file_path).expanduser().resolve()
        if not file_path.exists():
            raise ImageError(f"Invalid image at {file_path}: file does not exist.")
        media_type = mimetypes.guess_type(file_path)[0]
        if media_type not in VALID_MIMETYPES:
            raise ImageError(f"Invalid image type {media_type}: Image must be one of {VALID_MIMETYPES}.")
        with file_path.open("rb") as file:
            image_bytes = file.read()
        image = base64.b64encode(image_bytes).decode("utf-8")
        self.append(SpiceMessage(role="user", image_url=f"data:{media_type};base64,{image}", cache=cache))
        return self

    def add_prompt(self, role: Role, name: str, cache: bool = False, **context: Any) -> SpiceMessages:
        """Appends a message with the given role and pre-loaded prompt using jinja to render the context."""
        if self._client is None:
            raise ValueError("Cannot add prompt without a Spice client.")

        self.append(
            SpiceMessage(
                role=role,
                text=self._client.get_rendered_prompt(name, **context),
                cache=cache,
                prompt_metadata=PromptMetadata(name=name, content=self._client.get_prompt(name), context=context),
            )
        )
        return self

    def add_user_prompt(self, name: str, cache: bool = False, **context: Any) -> SpiceMessages:
        return self.add_prompt("user", name, cache, **context)

    def add_system_prompt(self, name: str, cache: bool = False, **context: Any) -> SpiceMessages:
        return self.add_prompt("system", name, cache, **context)

    def add_assistant_prompt(self, name: str, cache: bool = False, **context: Any) -> SpiceMessages:
        return self.add_prompt("assistant", name, cache, **context)

    def copy(self):
        new_copy = SpiceMessages(self._client, self)
        return new_copy



from typing import Any, Callable, Tuple

from spice.call_args import SpiceCallArgs
from spice.retry_strategy import Behavior, RetryStrategy
from spice.spice_message import SpiceMessages


def default_failure_message(message: str) -> str:
    return f"Failed to validate response for the following reason: {message}\n\nPlease try again."


class ValidatorStrategy(RetryStrategy):
    """
    Validates the model output and if it fails puts the failure reason in model context.

    """

    def __init__(
        self,
        validator: Callable[[str], Tuple[bool, str]],
        retries: int = 0,
        render_failure_message: Callable[[str], str] = default_failure_message,
    ):
        """
        Args:
            validator: A function that takes in the model output and returns a tuple of a boolean and a message. The
                boolean indicates if the model output is valid and the message is the reason why it is invalid.
            retries: The number of retries to attempt before failing.
            render_failure_message: A function that takes in the failure message and returns a string that will be
                displayed to the llm.
        """
        self.validator = validator
        self.retries = retries
        self.render_failure_message = render_failure_message

    def decide(
        self, call_args: SpiceCallArgs, attempt_number: int, model_output: str, name: str
    ) -> tuple[Behavior, SpiceCallArgs, Any, str]:
        passed, message = self.validator(model_output)
        if not passed:
            if attempt_number < self.retries:
                messages = SpiceMessages(messages=call_args.messages)
                messages.add_assistant_text(model_output)
                messages.add_user_text(self.render_failure_message(message))
                new_call_args = call_args.model_copy(update={"messages": messages})
                return Behavior.RETRY, new_call_args, None, f"{name}-retry-{attempt_number}-fail"
            else:
                raise ValueError("Failed to get a valid response after all retries")
        else:
            return Behavior.RETURN, call_args, model_output, name



from typing import Optional

from spice.models import EmbeddingModel, TextModel, TranscriptionModel


def text_request_cost(
    model: TextModel,
    input_tokens: int,
    cache_creation_input_tokens: int,
    cache_read_input_tokens: int,
    output_tokens: int,
) -> Optional[float]:
    if model.input_cost is not None and model.output_cost is not None:
        return (
            model.input_cost * input_tokens
            + 1.25 * model.input_cost * cache_creation_input_tokens
            + 0.10 * model.input_cost * cache_read_input_tokens
            + model.output_cost * output_tokens
        ) / 1_000_000
    else:
        return None


def embeddings_request_cost(model: EmbeddingModel, input_tokens: int) -> Optional[float]:
    if model.input_cost is not None:
        return (model.input_cost * input_tokens) / 1_000_000
    else:
        return None


def transcription_request_cost(model: TranscriptionModel, input_length: float) -> Optional[float]:
    if model.input_cost is not None:
        return (model.input_cost * input_length) / 100
    else:
        return None


def print_stream(text: str) -> None:
    print(text, end="", flush=True)


def string_identity(x: str) -> str:
    return x



from .spice import Spice, SpiceResponse, StreamingSpiceResponse, EmbeddingResponse, TranscriptionResponse  # noqa
from .utils import print_stream  # noqa
from .spice_message import SpiceMessages, SpiceMessage  # noqa



from abc import ABC, abstractmethod
from enum import Enum
from typing import Generic, TypeVar

from spice.call_args import SpiceCallArgs

T = TypeVar("T")

# An object that defines how get_response should validate/convert a response. It must implement the decide method.
# It takes in the previous call_args, which attempt number this is, and the model output.
# The method returns a tuple of Behavior, SpiceCallArgs, the result and the name of the run. If
# the Behavior is RETURN, then the result will be returned as the result of the SpiceResponse object. If the Behavior
# is RETRY, then the llm will be called again with the new spice_args. It's up to the Behavior to eventually return
# RETURN or throw an exception.


class Behavior(Enum):
    RETRY = "retry"
    RETURN = "return"


class RetryStrategy(ABC, Generic[T]):
    @abstractmethod
    def decide(
        self, call_args: SpiceCallArgs, attempt_number: int, model_output: str, name: str
    ) -> tuple[Behavior, SpiceCallArgs, T, str]:
        pass



import pytest

from spice import Spice
from spice.models import HAIKU
from spice.retry_strategy.converter_strategy import ConverterStrategy
from spice.retry_strategy.default_strategy import DefaultRetryStrategy
from spice.retry_strategy.validator_strategy import ValidatorStrategy
from tests.conftest import WrappedTestClient


@pytest.mark.asyncio
async def test_get_response_validator():
    client = WrappedTestClient(iter(["Hello, world!", "test"]))

    def equals_test(response):
        return response == "test"

    validator = equals_test

    spice = Spice()

    def return_wrapped_client(model, provider):
        return client

    spice._get_client = return_wrapped_client
    retry_strategy = DefaultRetryStrategy(validator=validator, retries=2)
    response = await spice.get_response(messages=[], model=HAIKU, retry_strategy=retry_strategy)

    assert response.text == "test"


@pytest.mark.asyncio
async def test_streaming_callback():
    client = WrappedTestClient(iter(["Hello, world!"]))

    spice = Spice()

    def return_wrapped_client(model, provider):
        return client

    spice._get_client = return_wrapped_client
    cache = ""

    def accumulator(text: str):
        nonlocal cache
        cache += text

    response = await spice.get_response(messages=[], model=HAIKU, streaming_callback=accumulator)

    assert response.text == "Hello, world!"
    assert cache == "Hello, world!"


@pytest.mark.asyncio
async def test_get_response_converter():
    client = WrappedTestClient(iter(["Not an int", "42"]))

    spice = Spice()

    def return_wrapped_client(model, provider):
        return client

    spice._get_client = return_wrapped_client
    retry_strategy = DefaultRetryStrategy(converter=int, retries=2)
    response = await spice.get_response(messages=[], model=HAIKU, retry_strategy=retry_strategy)

    assert response.text == "42"
    assert response.result == 42


@pytest.mark.asyncio
async def test_validator_retry_strategy():
    client = WrappedTestClient(iter(["Invalid response", "Valid response"]))

    def must_equal_valid(response):
        if response == "Valid response":
            return True, "You pass!"
        else:
            return False, "You fail!"

    retry_strategy = ValidatorStrategy(validator=must_equal_valid, retries=1)
    spice = Spice()

    def return_wrapped_client(model, provider):
        return client

    spice._get_client = return_wrapped_client
    response = await spice.get_response(messages=[], model=HAIKU, retry_strategy=retry_strategy)

    content = list(client.calls[-1].messages)[-1].content
    assert content.type == "text"
    last_message = content.text
    assert isinstance(last_message, str)
    assert "Failed to validate response for the following reason: You fail!" in last_message
    assert response.text == "Valid response"


@pytest.mark.asyncio
async def test_converter_retry_strategy():
    client = WrappedTestClient(iter(["Not an int", "42"]))

    spice = Spice()

    def return_wrapped_client(model, provider):
        return client

    spice._get_client = return_wrapped_client
    retry_strategy = ConverterStrategy(converter=int, retries=1)
    response = await spice.get_response(messages=[], model=HAIKU, retry_strategy=retry_strategy)

    content = list(client.calls[-1].messages)[-1].content
    assert content.type == "text"
    last_message = content.text
    assert isinstance(last_message, str)
    assert (
        "Failed to convert response for the following reason: invalid literal for int() with base 10: 'Not an int'"
        in last_message
    )
    assert response.text == "42"
    assert response.result == 42
